/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the mozCellML code.
 *
 * The Initial Developer of the Original Code is
 * The Bioengineering Institute, University of Auckland.
 * Portions created by the Initial Developer are Copyright (C) 2004
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Andrew Miller <ak.miller@auckland.ac.nz> (original developer)
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include "mozCellMLGraphClient.h"
#include "nsIDOMXULElement.h"
#include <nsIDOMDocument.h>
#include <nsIContent.h>
#include <nsIScriptGlobalObject.h>
#include <nsIDOMDocumentView.h>
#include <nsIPresShell.h>
#include <nsIPresContext.h>
#include <nsIDocShell.h>
#include <nsIFrame.h>
#include <nsIRenderingContext.h>
#include <nsIDOMAbstractView.h>
#include <nsUnitConversion.h>
#include <nsICategoryManager.h>
#include <nsIServiceManagerUtils.h>
#include "mozCellMLSimpleGraphProvider.h"

NS_IMPL_ISUPPORTS1(mozCellMLGraphClient, mozCellMLIGraphClient)

mozCellMLGraphClient::mozCellMLGraphClient()
{
}

mozCellMLGraphClient::~mozCellMLGraphClient()
{
}

NS_IMETHODIMP
mozCellMLGraphClient::Init(
                           double minBound,
                           double maxBound,
                           PRUint32 nvariables,
                           PRUint32 colours[],
                           double grid,
                           nsIDOMXULElement *el
                          )
{
  // TODO: Find a graph provider, and pass this to them.
  nsCOMPtr<nsICategoryManager> catman(do_GetService(NS_CATEGORYMANAGER_CONTRACTID));
  nsCOMPtr<nsISimpleEnumerator> cmi;
  catman->EnumerateCategory("mozcellml-graphproviders", getter_AddRefs(cmi));
  nsresult rv;
  while (true)
  {
    nsCOMPtr<nsISupports> gpSupports;
    rv = cmi->GetNext(getter_AddRefs(gpSupports));
    if (rv != NS_OK)
    {
      provider = new mozCellMLSimpleGraphProvider();
      break;
    }
    provider = do_QueryInterface(gpSupports, &rv);
    if (rv == NS_OK)
      break;
  }

  provider->Init(minBound, maxBound, nvariables, colours, grid);

  el->GetBoxObject(getter_AddRefs(mBoxObject));
  elem = el;

  Layout();
  Paint();
  return NS_OK;
}

NS_IMETHODIMP
mozCellMLGraphClient::Layout()
{
  return NS_OK;
}

NS_IMETHODIMP
mozCellMLGraphClient::Paint()
{
  // Now some messy stuff to get the nsIRenderingContext...
  nsresult rv;
  nsCOMPtr<nsIContent> content(do_QueryInterface(elem, &rv));
  NS_ENSURE_SUCCESS(rv, rv);
  nsCOMPtr<nsIDOMDocument> ownerDoc;
  rv = elem->GetOwnerDocument(getter_AddRefs(ownerDoc));
  NS_ENSURE_SUCCESS(rv, rv);
  nsCOMPtr<nsIDOMDocumentView> ownerView(do_QueryInterface(ownerDoc, &rv));
  NS_ENSURE_SUCCESS(rv, rv);
  nsCOMPtr<nsIDOMAbstractView> ownerAbstractView;
  rv = ownerView->GetDefaultView(getter_AddRefs(ownerAbstractView));
  NS_ENSURE_SUCCESS(rv, rv);
  nsCOMPtr<nsIScriptGlobalObject> so(do_QueryInterface(ownerAbstractView, &rv));
  NS_ENSURE_SUCCESS(rv, rv);
  nsCOMPtr<nsIPresShell> presShell;
  nsIDocShell* ds = so->GetDocShell();
  if (ds == nsnull)
    return NS_ERROR_FAILURE;
  rv = ds->GetPresShell(getter_AddRefs(presShell));
  NS_ENSURE_SUCCESS(rv, rv);
  nsIFrame *frame;
  presShell->GetPrimaryFrameFor(content, &frame);
  if (frame == nsnull)
    return NS_ERROR_FAILURE;
  nsCOMPtr<nsIRenderingContext> context;
  presShell->CreateRenderingContext(frame, getter_AddRefs(context));
  
  mBoxObject->GetX(&mX);
  mBoxObject->GetY(&mY);
  mBoxObject->GetWidth(&mWidth);
  mBoxObject->GetHeight(&mHeight);
  
  nsCOMPtr<nsIPresContext> pc;
  ds->GetPresContext(getter_AddRefs(pc));
  
  float p2t = pc->PixelsToTwips();
  mX = NSIntPixelsToTwips(mX, p2t);
  mY = NSIntPixelsToTwips(mY, p2t);
  mWidth = NSIntPixelsToTwips(mWidth, p2t);
  mHeight = NSIntPixelsToTwips(mHeight, p2t);
  if (provider != nsnull)
    return provider->Paint(context, mX, mY, mWidth, mHeight);
  return NS_OK;
}

NS_IMETHODIMP
mozCellMLGraphClient::SetColour(
                                PRUint32 traceId,
                                PRUint32 newColour
                               )
{
  if (provider != nsnull)
    provider->SetColour(traceId, newColour);
  /* Paint(); */
  return NS_OK;
}

NS_IMETHODIMP
mozCellMLGraphClient::SetVisibility(
                                    PRUint32 traceId,
                                    PRBool newVisibility
                                   )
{
  if (provider != nsnull)
    provider->SetVisibility(traceId, newVisibility);
  /* Paint(); */
  return NS_OK;
}

NS_IMETHODIMP
mozCellMLGraphClient::AddDataPoint(
                                   PRUint32 nvariables,
                                   double datapoint[]
                                  )
{
  if (provider != nsnull)
    provider->AddDataPoint(nvariables, datapoint);
  /* Paint(); */
  return NS_OK;
}
