const DATA_NODE = 1;
const AXES_NODE_BASE = 100000;

function GraphScaleMonitor(model)
{
  this.model = model;
  this.monitors = [];
  return this;
}

GraphScaleMonitor.prototype =
{
 QueryInterface: function(aIID)
 {
    if (
        aIID.equals(Components.interfaces.mozICellMLScaleWatcher) ||
        aIID.equals(Components.interfaces.nsISupports)
       )
    {
      return this;
    }
    throw Components.results.NS_NOINTERFACE;
 },

 addMonitor: function(m)
 {
   if (this.monitors != null)
     this.monitors.push(m);
 },

 removeMonitor: function(m)
 {
   if (this.monitors != null)
   {
     for (i = 0; i < this.monitors.length; i++)
       if (this.monitors[i] == m)
         delete this.monitors[i];
   }
 },

 scaleChanged: function (variable, newScale)
 {
   for (m in this.monitors)
   {
     m = this.monitors[m];
     m.scaleChanged(variable, newScale);
   }
 },

 progress: function(percentDone)
 {
   for (m in this.monitors)
   {
     m = this.monitors[m];
     m.progress(percentDone);
   }
 },

 done: function()
 {
   for (m in this.monitors)
   {
     m = this.monitors[m];
     m.done();
   }
   this.model = null;
   this.monitors = null;
 },

 halted: function ()
 {
   for (m in this.monitors)
   {
     m = this.monitors[m];
     m.halted();
   }
 },

 failed: function (msg)
 {
   for (m in this.monitors)
   {
     m = this.monitors[m];
     m.failed(msg);
   }
 }
};

function CmguiWrapper(m, cmgui)
{
  this.cellml = m.xmlmi.cellml.QueryInterface(Components.interfaces.mozICellMLLocal);
  this.cmgui = cmgui;
  this.plugins = [];
  this.lastSceneID = 0;
  return this;
}

function CORBARunner(f)
{
  this.run = f;
  return;
}

CORBARunner.prototype =
{
 QueryInterface: function(aIID)
 {
    if (
        aIID.equals(Components.interfaces.nsIRunnable) ||
        aIID.equals(Components.interfaces.nsISupports)
       )
    {
      return this;
    }
    throw Components.results.NS_NOINTERFACE;
 }
};

CmguiWrapper.prototype =
{
  ensure: function(f)
  {
    var cr = new CORBARunner(f);
    cr.ret = null;
    this.cellml.uiSafeRun(cr);
    return cr.ret;
  },

  begin: function()
  {
    if (this.inTransaction)
      this.commit();
    this.inTransaction = true;
    this.allCommands = "";
  },

  commit: function()
  {
    if (!this.inTransaction)
      return;
    this.inTransaction = false;
    var ac = this.allCommands;
    var cm = this.cmgui;
    this.ensure(function(){cm.executeCommand(ac);});
    this.allCommands = "";
  },

  rollback: function()
  {
    this.inTransaction = false;
    this.allCommands = "";
  },

  executeCommand: function(cmd)
  {
    var cleanCmd = "";
    for (i = 0; i < cmd.length; i++)
      if (cmd[i] == '\\')
        cleanCmd += '\\\\';
      else if (cmd[i] == '"')
        cleanCmd += '\\"';
      else
        cleanCmd += cmd[i];
    var wrapped  = "cmiss::cmiss(\"" + cleanCmd + "\");\n";

    if (this.inTransaction)
      this.allCommands += wrapped;
    else
    {
      var cm = this.cmgui;
      this.ensure(function(){cm.executeCommand(wrapped);});
    }
  },

  createStandardInterpolation: function(region, n, ar, c)
  {
    return this.ensure(function(){this.ret = region.createStandardInterpolation(n, ar, c);});
  },

  createTimeSequencePackage: function()
  {
    var cm = this.cmgui;
    return this.ensure(function(){this.ret = cm.createTimeSequencePackage()});
  },

  getSimpleTimeSequence: function(npoints, startx, gridx)
  {
    var tsp = this.timeSequencePackage;
    return this.ensure(function(){this.ret = tsp.getSimpleTimeSequence(npoints, startx, gridx);});
  },

  getMatchingTimeSequence: function(arr, npoints)
  {
    var tsp = this.timeSequencePackage;
    return this.ensure(function(){this.ret = tsp.getMatchingTimeSequence(arr, npoints);});
  },

  createRootNode: function(id)
  {
    var r = this.region;
    return this.ensure(function(){this.ret = r.createNode(id); });
  },

  createRootElement: function(id, dim)
  {
    var r = this.region;
    return this.ensure(function(){this.ret = r.createElementWithLineShape(id, dim); });
  },

  defineFieldOnNode: function(field, node, timeseq, creator)
  {
    this.ensure(function(){field.defineOnNode(node, timeseq, creator);});
  },

  createNode: function(parnode, id)
  {
    return this.ensure(function(){this.ret = parnode.createNode(id);});
  },

  createNodeFieldCreator: function(count)
  {
    var cm = this.cmgui;
    return this.ensure(function(){this.ret = cm.createNodeFieldCreator(count);});
  },

  createSceneViewer: function(handle, monitor)
  {
    var cm = this.cmgui;
    // createSceneViewer is oneway, no need to ensure.
    return cm.createSceneViewer
        (
         handle,
         ProxyToMain(
                     new SceneViewerMonitor(monitor),
                     Components.interfaces.CmguiISceneViewerResultMonitor
                    )
	);
  },

  getNodalValues: function(coords, node)
  {
    return this.ensure(function()
    {
      this.ret = coords.getNodalValues("", 0, node,
                                   Components.interfaces.CmguiIFunction.NODAL_VALUE,
                                   0, null);
    }
                      );
  },

  createMatrix: function(cols, data)
  {
    var cd = this.cmgui;
    return this.ensure(function()
    {
      this.ret = cd.createMatrix(cols, data, data.length);
    }
                      );
  },

  setValue: function(fv, m)
  {
    this.ensure(function()
    {
      fv.setValue(m);
    }
               );
  },

  setNodalValues: function(coords, node, cols, matrix)
  {
    this.setValue(this.getNodalValues(coords, node),
                  this.createMatrix(cols, matrix));
  },

  mergeNodeIntoRoot: function(node)
  {
    var rr = this.region;
    this.ensure(function(){rr.mergeNode(node);});
  },

  mergeElementIntoRoot: function(el)
  {
    var rr = this.region;
    this.ensure(function(){rr.mergeElement(el);});
  },

  defineTensorProductBasis: function(field, element, dimension, basis_type)
  {
    this.ensure(function()
    {
      field.defineTensorProductBasisOnElement(element, dimension, basis_type);
    }
               );
  },

  setElementNode: function(el, which, node)
  {
    this.ensure(function()
    {
      el.setNode(which, node);
    }
               );
  },

  findFieldByName: function(name)
  {
    var cm = this.cmgui;
    return this.ensure(function()
    {
      this.ret = cm.findFieldByName(name);
    }
                      );
  },

  createWindowProjectionField: function(name, sourceField, sViewer, projType)
  {
    var cm = this.cmgui;
    return this.ensure(function()
    {
      this.ret = cm.createWindowProjectionField(name, sourceField, sViewer, projType);
    }
                      );
  }
};

function SceneViewerWrapper(model, cd, plugin, w)
{
  this.w = w;
  this.m = model;
  this.cd = cd;
  this.plugin = plugin;
  cd.plugins.push(plugin);

  // Invent a scene name...
  this.sceneID = this.cd.lastSceneID + 1;
  this.cd.lastSceneID = this.sceneID;
  this.sceneName = "scene" + this.sceneID;

  // Create a new scene...
  this.cd.begin();
  this.cd.executeCommand("gfx create scene " + this.sceneName + " manual;");
  this.cd.executeCommand("gfx modify scene " + this.sceneName +
                         " add_light default;");
  this.cd.commit();

  // Create node for the axes...
  var axes = this.cd.createRootNode(AXES_NODE_BASE + this.sceneID);
  this.cd.mergeNodeIntoRoot(axes);
  this.cd.defineFieldOnNode(this.cd.coords, axes, null,
                            this.cd.coordFieldCreator);

  this.cd.setNodalValues(this.cd.coords, axes, 3, [-0.9, -0.9, 0]);
  this.cd.begin();
  this.cd.executeCommand("gfx create ngroup " + this.sceneName + ":axes;");
  this.cd.executeCommand("gfx modify ngroup " + this.sceneName + ":axes " +
                         " add " + (AXES_NODE_BASE + this.sceneID) + ";");
  this.cd.executeCommand("gfx draw scene " + this.sceneName + " group " +
                         this.sceneName + ":axes");
  this.cd.executeCommand("gfx modify g_element " + this.sceneName +
                         ":axes general clear " +
                         "circle_discretization 6 default_coordinate " +
                         "ndc_coordinates element_discretization \"4*4*4\"" +
                         " native_discretization none " + "scene " +
                         this.sceneName + ";");

  this.cd.commit();

  return this;
}

function SceneViewerMonitor(svw)
{
  this.wrapper = svw;
}

SceneViewerMonitor.prototype =
{
  QueryInterface: function(aIID)
  {
    if (
        aIID.equals(Components.interfaces.CmguiISceneViewerResultMonitor) ||
        aIID.equals(Components.interfaces.nsISupports)
       )
    {
      return this;
    }
    throw Components.results.NS_NOINTERFACE;
  },

  sceneViewerFailed: function()
  {
    this.wrapper.sceneViewerFailed();;
    this.wrapper = null;
  },

  sceneViewerReady: function(sv)
  {
    this.wrapper.sceneViewerReady(sv);
    this.wrapper = null;
  }
};

SceneViewerWrapper.prototype =
{
  QueryInterface: function(aIID)
  {
    if (
        aIID.equals(Components.interfaces.CmguiISceneViewerResultMonitor) ||
        aIID.equals(Components.interfaces.nsISupports)
       )
    {
      return this;
    }
    throw Components.results.NS_NOINTERFACE;
  },

  sceneViewerFailed: function()
  {
    alert("Error: Could not create a scene viewer.");
  },

  sceneViewerReady: function(sv)
  {
    this.sceneViewer = sv;
    this.setSceneByName(this.sceneName);
    this.setBackgroundColour(0.8, 0.8, 0.8);
    this.setViewportMode(Components.interfaces.CmguiISceneViewer.VIEWPORT_DISTORTING_RELATIVE);
    this.setTumbleRate(0);

    var rm = ProxyToMain(new NotifyingResizeMonitor(this),
                         Components.interfaces.mozIPluginResizeMonitor);

    this.plugin.resizeMonitor = rm;
    this.setWindowSize(this.plugin.width, this.plugin.height);

    /*
     * Now we have to set up our view. The graph rectangle lies in the XY
     * plane, with corners at (0,0,0), (1000,0,0), (1000,1000,0) and (0,1000,0).
     * The lookat point is in the middle of this rectangle at (500,500,0).
     * The eye point is then at (500,500,-10), with the up vector as (0,0,1).
     */
    /* this.setLookatParameters(500, 500, 500000, 500, 500, 0, 0, -1, 0); */

    // Get hold of the system coordinates field...
    var sc = this.cd.findFieldByName("system:coordinates");

    this.cd.createWindowProjectionField
      (this.sceneName + ":unscaledLabels", sc, sv,
       Components.interfaces.CmguiIComputedField.INVERSE_NDC_PROJECTION);

    // This is just the inverse of the normal scaling transformation...
    this.cd.begin();
    this.cd.executeCommand("gfx define field " + this.sceneName +
                           ":labels scale field " + this.sceneName +
                           ":unscaledLabels scale_factors 1 1 1;");

    this.cd.executeCommand("gfx modify g_element " + this.sceneName +
                           ":axes node_points as xaxis " +
                           "glyph line_ticks general size \"0*10*10\" centre " +
                           "0,0,0 " +
                           "label " + this.sceneName + ":labels.1 " +
                           "orientation xaxis " +
                           "scale_factors \"1.8*0*0\" select_on material " +
                           "system:black selected_material system:black " +
                           "scene " + this.sceneName + ";");
    this.cd.executeCommand("gfx modify g_element " + this.sceneName + ":axes " +
                           "node_points as yaxis " +
                           "glyph line_ticks general size \"0*1*1\" centre 0,0,0 " +
                           "label " + this.sceneName + ":labels.2 " +
                           "orientation yaxis " +
                           "scale_factors \"1.8*0*0\" select_on material " +
                           "system:black selected_material system:black " +
                           "scene " + this.sceneName + ";");
    this.cd.commit();

    // We want parallel projection...
    this.setProjectionMode(Components.interfaces.CmguiISceneViewer.PROJECTION_PARALLEL);

    /*
     * We want the graph rectangle to be just in view. The eye is distance 1000
     * from the graph plane, and the graph plane is(orthogonal to this distance)
     * 1000 wide. Therefore, we want a viewing angle of 2*arctan(500/1000)
     *  = 0.92730 rad
     */
    // this.setViewAngle(3.14159);

    if (this.w.activeGraph != null)
    {
      this.drawGraph(this.w.activeGraph);
    }
  },

  setSceneByName: function(name)
  {
    var sv = this.sceneViewer;
    this.cd.ensure(function(){sv.setSceneByName(name);});
  },

  setBackgroundColour: function(r, g, b)
  {
    var sv = this.sceneViewer;
    this.cd.ensure(function(){sv.setBackgroundColourRGB(r, g, b);});
  },

  viewAll: function()
  {
    var sv = this.sceneViewer;
    this.cd.ensure(function(){sv.viewAll();});
  },

  setViewportMode: function(x)
  {
    var sv = this.sceneViewer;
    this.cd.ensure(function(){sv.viewportMode = x;});
  },

  setTumbleRate: function(x)
  {
    var sv = this.sceneViewer;
    this.cd.ensure(function(){sv.tumbleRate = x;});
  },

  setWindowSize: function(w, h)
  {
    var sv = this.sceneViewer;
    sv.setWindowSize(w, h);
  },

  setProjectionMode: function(pm)
  {
    var sv = this.sceneViewer;
    this.cd.ensure(function(){sv.projectionMode = pm;});
  },

  setLookatParameters: function(ex,ey,ez,lx,ly,lz,ux,uy,uz)
  {
    var sv = this.sceneViewer;
    this.cd.ensure(function()
    {
      sv.setLookatParametersNonSkew(ex,ey,ez,lx,ly,lz,ux,uy,uz);
    }
                  );
  },

  getLookatParameters: function(ex,ey,ez,lx,ly,lz,ux,uy,uz)
  {
    var sv = this.sceneViewer;
    this.cd.ensure(function()
    {
      sv.getLookatParameters(ex,ey,ez,lx,ly,lz,ux,uy,uz);
    }
                  );
  },

  getViewingVolume: function(l, r, b, t, n, f)
  {
    var sv = this.sceneViewer;
    this.cd.ensure(function()
    {
      sv.getViewingVolume(l, r, b, t, n, f);
    }
                  );
  },

  setViewingVolume: function(l, r, b, t, n, f)
  {
    var sv = this.sceneViewer;
    this.cd.ensure(function()
    {
      sv.setViewingVolume(l, r, b, t, n, f);
    }
                  );
  },

  setViewAngle: function(a)
  {
    var sv = this.sceneViewer;
    this.cd.ensure(function()
    {
      sv.viewAngle = a;
    }
                  );
  },

  drawGraph: function(g)
  {
    this.showingGraph = g;

    RestoreGraphViewState(this, g);

    // Make a composite for every trace...
    var tid;
    this.w.xscale = 0;
    this.w.yscale = 0;
    this.cd.begin();
    for (tid in g.traces)
    {
      var t = g.traces[tid];
      var xfield = t.xvarlev.getFieldAgainstXI(this.m);
      var yfield = t.yvarlev.getFieldAgainstXI(this.m);
      this.cd.executeCommand("gfx define field " + this.sceneName + ":trace" + tid +
                             ":comp composite " + xfield + " " + yfield + " 0;");

      // We now have a field which provides points in 3D space for 2 variables,
      // varying across xi.1. They are in original units, so we have to scale
      // them to the pooled scale.
      var cv = t.xvarlev.cellmlvar;
      var xs = this.m.scales[cv.importNumber + ":" + cv.componentName + ":" + cv.name];
      cv = t.yvarlev.cellmlvar;
      var ys = this.m.scales[cv.importNumber + ":" + cv.componentName + ":" + cv.name];
      if (xs > this.w.xscale)
        this.w.xscale = xs;
      if (ys > this.w.yscale)
        this.w.yscale = ys;
    }
    this.cd.commit();

    // We have now determined xscale and yscale. Define the scales..
    RedefineGraphScale(this.w, t);

    // Finally, actually put the data on the graph...

    for (tid in g.traces)
    {
      var tn = this.sceneName + ":trace" + tid;
      var t = g.traces[tid];
      this.cd.begin();
      this.cd.executeCommand("gfx create egroup " + tn + ";");
      this.cd.executeCommand("gfx draw group " + tn + " scene " + this.sceneName);
      this.cd.executeCommand("gfx modify egroup " + tn + " add " + DATA_NODE + ";");
      // this.cd.executeCommand("gfx modify ngroup " + tn + " add " + DATA_NODE);
      var scaledField = this.sceneName + ":trace" + tid + ":scaled";
      // this.cd.executeCommand("gfx define field " + scaledField + " constant 0 0 0;");
      this.cd.executeCommand("gfx modify g_element " + tn + " general clear" +
                             " circle_discretization 6" +
                             " default_coordinate system:coordinates " +
                             " element_discretization \"" + (this.m.npoints-1) + "*1*1\"" +
                             " native_discretization none scene " + this.sceneName);
      // Make the new material...
      var rc, gc, bc;
      rc = ((t.colour >> 16) & 0xFF) / 255.0;
      gc = ((t.colour >> 8) & 0xFF) / 255.0;
      bc = (t.colour & 0xFF) / 255.0;
      this.cd.executeCommand("gfx create material " + tn + ":material diffuse " +
                             rc + " " + gc + " " + bc + " shininess 0");
      this.cd.executeCommand("gfx modify g_element " + tn + " element_points" +
                             " glyph point general size \"0.01*0.01*0.01\"" +
                             " coordinate " + scaledField +
                             " select_on " +
                             " discretization \"" + (this.m.npoints - 1) + "*1*1\"" +
                             " cell_corners" +
                             " material " + tn + ":material " +
                             " selected_material " + tn + ":material " +
                             /* " line_width 1" + */
                             " scene " + this.sceneName + ";");
      this.cd.commit();
    }
  },

  changeTraceColour: function(traceName, traceColour)
  {
    var r, g, b;
    r = ((traceColour >> 16) & 0xFF) / 255.0;
    g = ((traceColour >> 8) & 0xFF) / 255.0;
    b = (traceColour & 0xFF) / 255.0;
    this.cd.executeCommand("gfx modify material " + traceName +
                           ":material diffuse " + r + " " + g + " " + b +
                           " shininess 0");
  }
};

function RedefineGraphScale(vwin)
{
  var sv = vwin.sceneViewer;
  var g = vwin.activeGraph
  if (sv == null || g == null)
    return;
  var exs = vwin.xscale;
  var eys = vwin.yscale;
  if (exs == 0)
    exs = 1;
  if (eys == 0)
    eys = 1;
  // Set the scale so (-1,1) includes all the data...
  exs = 1.0 / exs;
  eys = 1.0 / eys;
  for (tid in g.traces)
  {
    sv.cd.executeCommand("gfx define field " + sv.sceneName + ":trace" + tid +
                         ":scaled scale field " + sv.sceneName + ":trace" + tid +
                         ":comp scale_factors " + exs + " " +
                         eys + " 1");
  }
  // Now fix up the axis labels.
  sv.cd.executeCommand("gfx define field " + sv.sceneName +
                       ":labels scale field " + sv.sceneName +
                       ":unscaledLabels scale_factors " + (1.0/exs) + " " +
                       (1.0 / eys) + " 1");
}


function NotifyingResizeMonitor(svw)
{
  this.sceneViewer = svw;
  return this;
}

NotifyingResizeMonitor.prototype =
{
  QueryInterface: function(aIID)
  {
    if (
        aIID.equals(Components.interfaces.mozIPluginResizeMonitor) ||
        aIID.equals(Components.interfaces.nsISupports))
      return this;
    throw Components.results.NS_NOINTERFACE;
  },

  onResize: function(width, height)
  {
    if (this.sceneViewer == null)
      return;
    this.sceneViewer.setWindowSize(width, height);
  }
};

function DoCmguiOneOffSetup(m)
{
  m.cmgui = new CmguiWrapper(m, m.cmgui);
  m.cmgui.region = m.cmgui.ensure(function(){this.ret = m.cmgui.cmgui.rootRegion; });
  m.cmgui.timeSequencePackage = m.cmgui.createTimeSequencePackage();

  m.cmgui.executeCommand("gfx modify scene default manual;");

  var npoints = parseInt((m.stopx - m.startx)/m.gridx);
  if (m.startx + npoints * m.gridx <= m.stopx)
    npoints++;
  m.npoints = npoints;

  var singleTime = m.cmgui.getMatchingTimeSequence([m.startx], 1);
  m.cmgui.pointFieldCreator = m.cmgui.createNodeFieldCreator(1);
  m.cmgui.coordFieldCreator = m.cmgui.createNodeFieldCreator(3);

  m.cmgui.axesTemplate = m.cmgui.createRootNode(AXES_NODE_BASE);
  m.cmgui.dataNode = m.cmgui.createRootNode(DATA_NODE);
  m.cmgui.mergeNodeIntoRoot(m.cmgui.dataNode);

  var el = m.cmgui.createRootElement(DATA_NODE, 1);

  // Firstly, create a field that offsets from xi.1...
  m.cmgui.begin();
  var scalf = parseFloat(m.stopx)-parseFloat(m.startx);
  m.cmgui.executeCommand("gfx define field system:xiscale scale field " +
                         "xi.1 scale_factors " + scalf);
  m.cmgui.executeCommand("gfx define field system:xitobound offset field " + 
                         "system:xiscale offsets " + m.startx);
  m.cmgui.executeCommand("gfx set point_size 5");
  m.cmgui.commit();

  // Create one field for each variable...
  var i;

  m.cmgui.valuefield = [];
  for (i = 0; i < m.xmlmi.nivariables; i++)
  {
    var name = "field" + i;
    var varfield =
      m.cmgui.createStandardInterpolation(m.cmgui.region, name + ":raw",
                                          ["value"], 1);
    m.cmgui.valuefield[i] = varfield;
    m.cmgui.defineFieldOnNode(varfield,
                              m.cmgui.dataNode,
                              singleTime,
                              m.cmgui.pointFieldCreator);
    m.cmgui.executeCommand("gfx define field " + name + ":nlookup " +
                           "nodal_lookup field " + name + ":raw " +
                           "node " + DATA_NODE);

    // And a time lookup. This is needed because the graph drawing code holds
    // the time constant but varies the xi. xitobound contains a mapping from
    // these xi coordinates to times, which are looked up on nlookup.
    m.cmgui.executeCommand("gfx define field " + name + ":tlookup " +
                           "time_lookup field " + name + ":nlookup " +
                           "time_field system:xitobound");
  }

  // Create a co-ordinate field...
  m.cmgui.coords = m.cmgui.createStandardInterpolation(m.cmgui.region,
                                                       "system:coordinates",
                                                       ["x", "y", "z"], 3);

  // Define a tensor product basis on our element...
  m.cmgui.defineTensorProductBasis(m.cmgui.coords, el, 1,
                                   Components.interfaces.CmguiIFunction
                                             .BASIS_LINEAR_LAGRANGE);
  m.cmgui.mergeElementIntoRoot(el);

  m.cmgui.begin();
  // Define the co-ordinate system...
  m.cmgui.executeCommand("gfx define field ndc_coordinates coordinate_system " +
                         "normalised_window_coordinates composite " +
                         "system:coordinates;");
  // Define some constants for the axis code...
  m.cmgui.executeCommand("gfx define field xaxis constant 1 0 0;");
  m.cmgui.executeCommand("gfx define field yaxis constant 0 1 0;");
  // Define a black material for the axis...
  m.cmgui.executeCommand("gfx create material system:black " +
                         "diffuse 0 0 0 specular 0 0 0 shininess 0");
  m.cmgui.commit();
}

function PrepareNewScene(w, m, handle, width, height)
{
  if (m.graphs == null)
    m.graphs = [];
  if (m.graphs.length > 1)
  {
    w.activeGraph = m.graphs[1];
  }
  w.sceneViewer = new SceneViewerWrapper(m, m.cmgui, w.plugin, w);
  m.cmgui.createSceneViewer(handle, w.sceneViewer);
}


function DestroySingleTrace(w, traceid)
{
  if (w.sceneViewer != null)
  {
    var tn = w.sceneViewer.sceneName + ":trace" + traceid;
    var m = w.activeModel;
    m.cmgui.executeCommand("gfx destroy egroup " + tn);
    m.cmgui.executeCommand("gfx destroy material " + tn + ":material");
    m.cmgui.executeCommand("gfx destroy field " + tn + ":scaled");
    m.cmgui.executeCommand("gfx destroy field " + tn + ":comp");
  }
}

function DestroyGraph(w, g)
{
  // We don't destroy the axes...
  var i;
  for (i in g.traces)
    DestroySingleTrace(w, i);
}

function DestroyScene(w)
{
  if (w.activeModel == null || w.activeGraph == null)
    return;
  // Get rid of all the traces...
  DestroyGraph(w, w.activeGraph);
  var m = w.activeModel;
  m.cmgui.executeCommand("gfx erase scene " + w.sceneViewer.sceneName +
                         " group " + w.sceneViewer.sceneName + ":axes");
  m.cmgui.executeCommand("gfx destroy ngroup " + w.sceneViewer.sceneName +
                         ":axes");
  m.cmgui.executeCommand("gfx destroy nodes " +
                         (AXES_NODE_BASE + w.sceneViewer.sceneID));
  m.cmgui.executeCommand("gfx destroy field " + w.sceneViewer.sceneName +
                         ":labels.1");
  m.cmgui.executeCommand("gfx destroy field " + w.sceneViewer.sceneName +
                         ":labels.2");
  m.cmgui.executeCommand("gfx destroy field " + w.sceneViewer.sceneName +
                         ":labels");
  m.cmgui.executeCommand("gfx destroy field " + w.sceneViewer.sceneName +
                         ":unscaledLabels");
  m.cmgui.executeCommand("gfx destroy scene " + w.sceneViewer.sceneName);
  var p;
  for (p in m.cmgui.plugins)
  {
    if (m.cmgui.plugins[p] == w.sceneViewer.plugin)
      delete m.cmgui.plugins[p];
  }
}
