/**
 * Constructor for the ModelInstance class. A ModelInstance represents a
 * particular set of initial conditions, integration parameters, as well as the
 * mathematical structure of the document.
 * @param cellml The mozICellML containing the mathematical structure. This may
 *               be shared by several mozICellML objects.
 */
function ModelInstance(cellml, xmlmi)
{
  this.init(cellml, xmlmi);
  return this;
}

function FillString(s, z, l)
{
  s = s + "";
  while (s.length < l)
    s = z + s;
  return s;
}

function FormatDate(d, f)
{
  var i;
  var str = "";
  for (i = 0; i < f.length; i++)
  {
    switch (f[i])
    {
    case 'D':
      str += FillString(d.getDate(), "0", 2);
      break;
    case 'm':
      str += FillString(d.getMonth(), "0", 2);
      break;
    case 'Y':
      str += FillString(d.getYear(), "0", 2);
      break;
    case 'H':
      str += FillString(d.getHours(), "0", 2);
      break;
    case 'M':
      str += FillString(d.getMinutes(), "0", 2);
      break;
    case 'S':
      str += FillString(d.getSeconds(), "0", 2);
      break;
    default:
      str += f[i];
    }
  }
  return str;
}

/**
 * A partial mozICellMLVariable implementation which returns a particular initial value...
 */
function ModelVariableWrapper(iv)
{
  this.initialValue = iv;
  return this;
}

/**
 * The prototype for ModelVariableWrapper.
 */
ModelVariableWrapper.prototype =
{
  /**
   * A standard QueryInterface method. This particular class supports
   *  mozICellMLVariable and nsISupports.
   */
  QueryInterface: function(aIID)
  {
    if (
        aIID.equals(Components.interfaces.mozICellMLVariable) ||
        aIID.equals(Components.interfaces.nsISupports)
       )
    {
      return this;
    }
    throw Components.results.NS_NOINTERFACE;
  }
};

/**
 * The prototype for the ModelInstance class.
 */
ModelInstance.prototype =
{
  /**
   * Internal function called by the constructor to set-up a ModelInstance.
   */
  init: function(cellml, xmlmi)
  {
    this.dirty = false;
    this.results = null;
    this.startx = 0;
    this.stopx = 10;
    this.gridx = 1;
    this.maxStep = 1;
    this.relError = 1E-6;
    this.absError = 1E-6;
    this.cellml = cellml;
    this.xmlmi = xmlmi;
    var d = new Date();
    this.name = FormatDate(d, "D/m H:M:S");
    return this;
  },

  /**
   * A standard QueryInterface method. This particular class supports
   *  mozICellML and nsISupports.
   */
  QueryInterface: function(aIID)
  {
    if (
        aIID.equals(Components.interfaces.mozICellML) ||
        aIID.equals(Components.interfaces.nsISupports)
       )
    {
      return this;
    }
    throw Components.results.NS_NOINTERFACE;
  },

  getVariable: function(importNo, componentName, name)
  {
    var iv = 0.0;
    try
    {
      iv = this.getInitialValue(importNo, componentName, name);
    }
    catch (e)
    {
    }
    return new ModelVariableWrapper(iv);
  },

  /**
   * Adds this model to the model table on a given window.
   * @param w The window which holds a last_instance_id and model_table
   *        variable.
   * Side-effect Increments the last_instance_id of the window, and stores the
   *             new instance ID.
   */
  addToModelTable: function(w)
  {
    if (w.last_instance_id == null)
      w.last_instance_id = 0;
    this.win = w;
    if (this.instance_id == null)
      this.instance_id = w.last_instance_id++;
    if (w.model_table == null)
      w.model_table = [];
    w.model_table[this.instance_id] = this;
  },

  /**
   * Sets RDF service, components, and resources for this component.
   * @param serv The nsIRDFService.
   * @param name_resource The predicate to use for the name.
   * @param intanceid_resource The predicate to use for the instance ID.
   * @param source_latest The RDFSource for the latest model view.
   * @param source_flat The RDFSource for the flat model view.
   * @param source_nested The RDFSource for the nested model view.
   */
  setRDF: function(serv, name_resource, instanceid_resource, source_latest,
                   source_flat, source_nested, cont_utils)
  {
    this.rdfserv = serv;
    this.rdfname_resource = name_resource;
    this.rdfinstanceid_resource = instanceid_resource;
    this.rdfsource_latest = source_latest;
    this.rdfsource_flat = source_flat;
    this.rdfsource_nested = source_nested;
    this.rdfcontainer_utils = cont_utils;
  },

  /**
   * Adds this ModelInstance to every container.
   */
  addToAllContainers: function(container_latest, container_flat, container_nested)
  {
    this.rdfcontainer_latest = container_latest;
    this.rdfcontainer_flat = container_flat;
    this.rdfcontainer_nested = container_nested;
    this.xmlmi.latest = this;
    
    this.addToRDF(this.rdfsource_latest, this.rdfcontainer_latest);
    this.addToRDF(this.rdfsource_flat, this.rdfcontainer_flat);
    this.addToRDF(this.rdfsource_nested, this.rdfcontainer_nested);
  },

  /**
   * Called when a new ModelInstance takes over as the latest instance for a
   * given XMLModelInstance.
   */
  noLongerLatest: function()
  {
    this.rdfcontainer_latest.RemoveElement(this.self_resource, true);
  },

  /**
   * Produces a clone of this model and puts it beneath this model in the
   * change hierarchy.
   */
  cloneInstance: function()
  {
    var newNode = new ModelInstance(this.cellml, this.xmlmi);
    if (arguments.length > 0)
    {
      if (arguments[0] != null)
        newNode.name = arguments[0];
      if (arguments.length > 1 && arguments[1] != null)
        newNode.instance_id = arguments[1];
    }
    newNode.parentInstance = this;
    newNode.dirty = true;
    newNode.startx = this.startx;
    newNode.stopx = this.stopx;
    newNode.gridx = this.gridx;
    newNode.maxStep = this.maxStep;
    newNode.relError = this.relError;
    newNode.absError = this.absError;


    newNode.setRDF(this.rdfserv, this.rdfname_resource,
                   this.rdfinstanceid_resource, this.rdfsource_latest,
                   this.rdfsource_flat, this.rdfsource_nested,
                   this.rdfcontainer_utils);

    this.noLongerLatest();
    if (this.rdfcontainer_self == null)
    {
      this.rdfcontainer_self =
        this.rdfcontainer_utils.MakeSeq(this.rdfsource_nested, this.self_resource);
    }
    newNode.addToModelTable(this.win);
    newNode.addToAllContainers(this.rdfcontainer_latest, this.rdfcontainer_flat, this.rdfcontainer_self);

    newNode.initialValues = [];
    var i, imp, comp, varn;
    if (this.initialValues != null)
    {
      for (imp = 0; imp < this.initialValues.length; imp++)
        for (comp in this.initialValues[imp])
          for (varn in this.initialValues[imp][comp])
          {
            if (newNode.initialValues[imp] == null)
              newNode.initialValues[imp] = {};
            if (newNode.initialValues[imp][comp] == null)
              newNode.initialValues[imp][comp] = {};
            newNode.initialValues[imp][comp][varn] = {};
            newNode.initialValues[imp][comp][varn].value = this.initialValues[imp][comp][varn].value;
            newNode.initialValues[imp][comp][varn].monitors = [];
          }
    }
    else
    {
      var allVars = {};
      var len = {};
      newNode.cellml.getAllVariables(Components.interfaces.mozICellML.
                                     SHOULD_HAVE_INITIAL_VALUE, allVars, len);
      allVars = allVars.value;
      newNode.initialValues = [];
      for (i = 0; i < allVars.length; i++)
      {
        imp = allVars[i].importNumber;
        comp = allVars[i].componentName;
        varn = allVars[i].name;
        if (newNode.initialValues[imp] == null)
          newNode.initialValues[imp] = {};
        if (newNode.initialValues[imp][comp] == null)
          newNode.initialValues[imp][comp] = {};
        newNode.initialValues[imp][comp][varn] = {};
        try
        {
          newNode.initialValues[imp][comp][varn].value = allVars[i].initialValue;
        }
        catch (e)
        {
          newNode.initialValues[imp][comp][varn].value = 0.0;
        }
        newNode.initialValues[imp][comp][varn].monitors = [];
      }
    }
    return newNode;
  },

  /**
   * Sets up the RDF triples for this model on a particular source.
   * @param source The nsIRDFSource on which to set up the triples.
   * @param container The nsIRDFContainer into which to insert the model
   *                  resource.
   */
  addToRDF: function(source, container)
  {
    if (this.self_resource == null)
    {
      this.self_resource = this.rdfserv.GetAnonymousResource();
    }
    if (this.name_literal == null)
    {
      this.name_literal = this.rdfserv.GetLiteral(this.name);
    }
    if (this.instance_id_literal == null)
    {
      this.instance_id_literal = this.rdfserv.GetIntLiteral(this.instance_id);
    }
    source.Assert(this.self_resource, this.rdfname_resource, this.name_literal, true);
    source.Assert(this.self_resource, this.rdfinstanceid_resource, this.instance_id_literal, true);
    container.AppendElement(this.self_resource);
  },

  /**
   * Modifies the initial value of a variable.
   */
  setInitialValue: function(imp, comp, varn, iv)
  {
    if (this.initialValues == null)
    {
      alert("XMLModelInstances are supposed to be immutable, but " +
            "setInitialValue was just called! This is a bug.");
      throw Components.results.NS_ERROR_FAILURE;
    }

    this.initialValues[imp][comp][varn].value = iv;
    var m = this.initialValues[imp][comp][varn].monitors;
    var i;
    for (i = 0; i < m.length; i++)
      m[i].updateValue(iv);
  },

  /**
   * Retrieves the initial value of a variable.
   */
  getInitialValue: function(imp, comp, varn)
  {
    if (this.initialValues == null)
    {
      return this.cellml.getVariable(imp, comp, varn).initialValue;
    }
    else
      return this.initialValues[imp][comp][varn].value;
  },

  /**
   * Allows a VariableLevel to be informed when changes to the initial value of
   * a variable in this model occurs.
   * @param varlev The VariableLevel concerned.
   * Note This only holds a *weak* reference to the VariableLevel, so you need
   *      to reference it from somewhere else as well.
   */
  monitorInitialConditions: function(varlev)
  {
    var imp, comp, varn;
    varn = varlev.name;
    comp = varlev.complev.name;
    imp = varlev.implev.number;

    if (this.initialValues != null &&
        this.initialValues[imp] != null &&
        this.initialValues[imp][comp] != null &&
        this.initialValues[imp][comp][varn] != null)
    {
      this.initialValues[imp][comp][varn].monitors.push(varlev);
    }
  },

  /**
   * Cancels the notification registered with monitorInitialConditions.
   */
  removeMonitor: function(varlev)
  {
    var imp, comp, varn;
    varn = varlev.name;
    comp = varlev.complev.name;
    imp = varlev.implev.number;

    if (this.initialValues != null &&
        this.initialValues[imp] != null &&
        this.initialValues[imp][comp] != null &&
        this.initialValues[imp][comp][varn] != null)
    {
      var m = this.initialValues[imp][comp][varn].monitors;
      var i;
      for (i = 0; i < m.length; i++)
        if (m[i] == varlev)
          break;
      this.initialValues[imp][comp][varn].monitors = m.slice(0, i).concat(m.slice(i + 1, m.length));
    }
  },

  /**
   * Determines the number of variables in this model instance...
   */
  countVariables: function(varlev)
  {
    if (m.xmlmi.nvariables != null)
      return m.xmlmi.nvariables;
    m.getAllVariables();
  }
};


/**
 * The constructor for the XMLModelInstance class.
 * An XMLModelInstance is a special case of a ModelInstance where the model was
 * loaded from an XML file.
 * @param name The name of the model(can be derived from the filename).
 * @param cellml The mozICellML representing the mathematics and boundary conditions.
 */
function XMLModelInstance(name, cellml)
{
  this.init(cellml, this);
  this.name = name;

  var allvars = new Object();
  var numvars = new Object();
  cellml.getAllVariables(Components.interfaces.mozICellML.NO_IN_VARIABLES,
                         allvars, numvars);
  this.variables = allvars.value;
  this.nvariables = numvars.value;

  this.ivariables = [];
  this.nivariables = 0;
  var v;
  for (v in this.variables)
  {
    v = this.variables[v];
    var vt = v.type;
    if (vt == Components.interfaces.mozICellMLVariable.SIMPLE ||
        vt == Components.interfaces.mozICellMLVariable.DIFFERENTIAL)
    {
      this.ivariables.push(v);
      this.nivariables++;
    }
    else if (vt == Components.interfaces.mozICellMLVariable.BOUND)
      this.bvar = v;
  }

  return this;
}

/**
 * The prototype for XMLModelInstance.
 */
XMLModelInstance.prototype = new ModelInstance;

function AxisVariable(cellmlvar)
{
  this.impno = cellmlvar.importNumber;
  this.compName = cellmlvar.componentName;
  this.name = cellmlvar.name;
  this.cellmlvar = cellmlvar;
}

AxisVariable.prototype =
{
  getFieldAgainstXI: function(m)
  {
    var vt = this.cellmlvar.type;
    if (vt == Components.interfaces.mozICellMLVariable.BOUND)
      return "system:xitobound";
    else if (vt == Components.interfaces.mozICellMLVariable.CONSTANT)
      return ("" + this.cellmlvar.initialValue);
    // Otherwise, it is one of the "ivariables". Find the number...
    for (ivi in m.xmlmi.ivariables)
    {
      var iv = m.xmlmi.ivariables[ivi];
      if (iv == this.cellmlvar)
        return "field" + ivi + ":tlookup";
    }
    return "field_not_found_so_report_a_bug";
  }
};

/**
 * A constructor for the VariableLevel class.
 * VariableLevel classes represent the variable level of the model=>import=>
 * component=>variable hierarachy, and maintain RDF to mirror this information.
 * @param modlev The model level above this Variable
 * @param name The name of the variable.
 * @param cellmlvar The mozICellMLVariable instance.
 * @param idx The variableIndex to store in the RDF.
 * Side-effect This code will automatically add or find the correct import and
 *             component from the model level, and add itself into the appropriate
 *             RDF containers.
 */
function VariableLevel(modlev, name, cellmlvar, idx)
{
  this.name = name;
  this.cellmlvar = cellmlvar;
  this.modlev = modlev;
  this.resource = this.modlev.rdfserv.GetAnonymousResource();
  this.important = false;

  this.implev = this.modlev.addOrFindImport(cellmlvar.importNumber);
  this.complev = this.implev.addOrFindComponent(cellmlvar.componentName);

  this.complev.addVariable(name, this);

  var vnr = this.modlev.rdfserv.GetResource
    ("http://www.physiome.org/rdf/mozCellML#variableName");
  var vir = this.modlev.rdfserv.GetResource
    ("http://www.physiome.org/rdf/mozCellML#variableIndex");
  var vtr = this.modlev.rdfserv.GetResource
    ("http://www.physiome.org/rdf/mozCellML#variableType");
  var vvr = this.modlev.rdfserv.GetResource
    ("http://www.physiome.org/rdf/mozCellML#variableValue");
  var vur = this.modlev.rdfserv.GetResource
    ("http://www.physiome.org/rdf/mozCellML#variableUnits");

  var value = "";
  try
  {
    value = this.modlev.model.getInitialValue(this.implev.number,
                                              this.complev.name,
                                              this.name);
  }
  catch (e)
  {
  }

  this.value = value;

  var vnl = this.modlev.rdfserv.GetLiteral(name);
  var vil = this.modlev.rdfserv.GetIntLiteral(idx);
  var vtl = this.modlev.rdfserv.GetLiteral(TypeToString(cellmlvar.type));
  var vvl = this.modlev.rdfserv.GetLiteral(value);
  var vul = this.modlev.rdfserv.GetLiteral(cellmlvar.units);

  this.modlev.rsallt.Assert(this.resource, vnr, vnl, true);
  this.modlev.rsimportantt.Assert(this.resource, vnr, vnl, true);
  this.modlev.rsallf.Assert(this.resource, vnr, vnl, true);
  this.modlev.rsimportantf.Assert(this.resource, vnr, vnl, true);

  this.modlev.rsallt.Assert(this.resource, vir, vil, true);
  this.modlev.rsimportantt.Assert(this.resource, vir, vil, true);
  this.modlev.rsallf.Assert(this.resource, vir, vil, true);
  this.modlev.rsimportantf.Assert(this.resource, vir, vil, true);

  this.modlev.rsallt.Assert(this.resource, vtr, vtl, true);
  this.modlev.rsimportantt.Assert(this.resource, vtr, vtl, true);
  this.modlev.rsallf.Assert(this.resource, vtr, vtl, true);
  this.modlev.rsimportantf.Assert(this.resource, vtr, vtl, true);

  this.modlev.rsallt.Assert(this.resource, vvr, vvl, true);
  this.modlev.rsimportantt.Assert(this.resource, vvr, vvl, true);
  this.modlev.rsallf.Assert(this.resource, vvr, vvl, true);
  this.modlev.rsimportantf.Assert(this.resource, vvr, vvl, true);

  this.modlev.rsallt.Assert(this.resource, vur, vul, true);
  this.modlev.rsimportantt.Assert(this.resource, vur, vul, true);
  this.modlev.rsallf.Assert(this.resource, vur, vul, true);
  this.modlev.rsimportantf.Assert(this.resource, vur, vul, true);

  this.modlev.rcallf.AppendElement(this.resource);

  this.modlev.model.monitorInitialConditions(this);
  return this;
}

/**
 * The prototype for the VariableLevel class.
 */
VariableLevel.prototype = {
  /**
   * Set/clear the important flag on this variable.
   * @param important True if this variable is now marked as important, or false
   *        otherwise.
   */
  setImportant: function(important)
  {
    if (this.important == important)
      return;

    if (important)
      this.modlev.rcimportantf.AppendElement(this.resource);
    else
      this.modlev.rcimportantf.RemoveElement(this.resource, true);

    this.complev.setVariableImportant(this.name, important);
  },

  /**
   * Called by the model whenever the value has been updated.
   * @param newval A variable containing the value.
   */
  updateValue: function(newval)
  {
    var vvlo = this.modlev.rdfserv.GetLiteral(this.value);
    this.value = newval;
    var vvln = this.modlev.rdfserv.GetLiteral(this.value);
    var vvr = this.modlev.rdfserv.GetResource
      ("http://www.physiome.org/rdf/mozCellML#variableValue");
    
    this.modlev.rsallt.Unassert(this.resource, vvr, vvlo);
    this.modlev.rsimportantt.Unassert(this.resource, vvr, vvlo);
    this.modlev.rsallf.Unassert(this.resource, vvr, vvlo);
    this.modlev.rsimportantf.Unassert(this.resource, vvr, vvlo);

    this.modlev.rsallt.Assert(this.resource, vvr, vvln, true);
    this.modlev.rsimportantt.Unassert(this.resource, vvr, vvln, true);
    this.modlev.rsallf.Assert(this.resource, vvr, vvln, true);
    this.modlev.rsimportantf.Assert(this.resource, vvr, vvln, true);

    //if (this.modlev.tree != null)
    //  this.modlev.tree.builder.rebuild();
  },

  destroy: function()
  {
    this.modlev.model.removeMonitor(this);
  },

  selectInTree: function(t)
  {
    try
    {
      var builder = t.view.QueryInterface(Components.interfaces.nsIXULTreeBuilder);
      var treeidx = builder.getIndexOfResource(this.resource);
      t.view.selection.clearSelection();
      t.view.selection.currentIndex = treeidx;
      t.view.selection.select(treeidx);
    }
    catch (e)
    {
       return -1;
    }
  }
};

/**
 * A constructor for the ComponentLevel class.
 * ComponentLevel classes represent the component level of the model=>import=>
 * component=>variable hierarachy, and maintain RDF to mirror this information.
 * @param name The name of this component.
 * @param modlev The model level above this component.
 * @param implev The import level above this component.
 * Side-effect This code will automatically add itself into the appropriate
 *             RDF containers.
 * Note You should probably call ImportLevel::addOrFindComponent instead of
 *      calling the constructor yourself.
 */
function ComponentLevel(name, modlev, implev)
{
  this.modlev = modlev;
  this.implev = implev;
  this.name = name;
  this.variables = {};

  this.resource = this.modlev.rdfserv.GetAnonymousResource();

  this.modlev.rsallt.Assert
  (
    this.resource,
    this.modlev.rdfserv.GetResource
    (
     "http://www.physiome.org/rdf/mozCellML#variableName"
    ),
    this.modlev.rdfserv.GetLiteral(name), true
  );
  this.modlev.rsimportantt.Assert
  (
    this.resource,
    this.modlev.rdfserv.GetResource
    (
     "http://www.physiome.org/rdf/mozCellML#variableName"
    ),
    this.modlev.rdfserv.GetLiteral(name), true
  );

  this.rcallt = this.modlev.rdfcu.MakeSeq(this.modlev.rsallt, this.resource);
  this.rcimportantt = this.modlev.rdfcu.MakeSeq(this.modlev.rsimportantt, this.resource);

  return this;
}

/**
 * The prototype for the ComponentLevel class.
 */
ComponentLevel.prototype = {
  /**
   * Readonly outside of this class.
   * The number of important variables in this component.
   */
  importantVariables: 0,
  /**
   * Readonly outside of this class.
   * The set of all VariableLevels in this ComponentLevel.
   */
  variables: {},
  /**
   * Readonly outside of this class.
   * The number of variables in this component.
   */
  nvariables: 0,

  /**
   * Adds a variable to this component. Called automatically by the variable
   * constructor, do not call directly.
   * @param name The name of the variable.
   * @param v The VariableLevel to add.
   */
  addVariable: function (name, v)
  {
    this.variables[name] = v;
    this.nvariables++;
    this.rcallt.AppendElement(v.resource);
  },

  /**
   * Sets a variable as important. Called automatically by
   * VariableLevel::setImportant, do not call directly.
   */
  setVariableImportant: function(name, isimportant)
  {
    if (this.variables[name].important == isimportant)
      return;

    this.variables[name].important = isimportant;

    if (this.importantVariables == 0 && isimportant)
      this.implev.setComponentImportant(this.name, true);
    else if (this.importantVariables == 1 && !isimportant)
      this.implev.setComponentImportant(this.name, false);

    if (isimportant)
    {
      this.importantVariables++;
      this.rcimportantt.AppendElement(this.variables[name].resource);
    }
    else
    {
      this.importantVariables--;
      this.rcimportantt.RemoveElement(this.variables[name].resource, true);
    }
  },

  stringifyImportantState: function()
  {
    var k;
    var str = "";
    for (k in this.variables)
      if (this.variables[k].important)
      {
        str = str + k + " ";
      }
    return str;
  },

  destroy: function()
  {
    var x;
    for (x in this.variables)
    {
      this.variables[x].destroy();
      this.variables[x] = null;
    }
  }
};

/**
 * A constructor for the ImportLevel class.
 * ImportLevel classes represent the import level of the model=>import=>
 * component=>variable hierarachy, and maintain RDF to mirror this information.
 * @param number The number assigned to this import.
 * @param modlev The model level above this import.
 * Side-effect This code will automatically add itself into the appropriate
 *             RDF containers.
 * Note You probably should call ModelLevel::addOrFindImport instead of calling
 *      the constructor yourself.
 */
function ImportLevel(number, modlev)
{
  this.number = number;
  this.modlev = modlev;
  this.components = {};

  this.resource = this.modlev.rdfserv.GetAnonymousResource();

  this.name = this.modlev.model.cellml.describeImportNumber(number);

  var nl = this.modlev.rdfserv.GetLiteral(this.name);
  this.modlev.rsallt.Assert(this.resource,
                           this.modlev.rdfserv.GetResource
                           ("http://www.physiome.org/rdf/mozCellML#variableName"),
                           nl, true);
  this.modlev.rsimportantt.Assert(this.resource,
                     this.modlev.rdfserv.GetResource
                     ("http://www.physiome.org/rdf/mozCellML#variableName"),
                     nl, true);

  this.rcallt = this.modlev.rdfcu.MakeSeq(this.modlev.rsallt, this.resource);
  this.rcimportantt = this.modlev.rdfcu.MakeSeq(this.modlev.rsimportantt, this.resource);

  return this;
};

/**
 * The prototype for the ImportLevel class.
 */
ImportLevel.prototype = {
  /**
   * Readonly outside of this class.
   * The set of all components in this import level.
   */
  components: {},

  /**
   * Readonly outside of this class.
   * The number of components which contain one or more important variables.
   */
  importantComponents: 0,

  /**
   * Readonly outside of this class.
   * The number of non-empty components.
   */
  nonemptyComponents: 0,

  /**
   * Tries to find a component, and if it doesn't exist, makes a new one.
   * @param name The name of the component to try to find.
   * @return The existing or new component.
   */
  addOrFindComponent: function(name)
  {
    if (this.components[name] == null)
    {
      this.nonemptyComponents++;
      this.components[name] = new ComponentLevel(name, this.modlev, this);
      this.rcallt.AppendElement(this.components[name].resource);
    }
    return this.components[name];
  },

  /**
   * Sets a component as containing(or not containing) an important variable.
   * This is called from ComponentLevel::setVariableImportant, do not call
   * directly.
   * @param name The name of the component.
   * @param isimportant True if the component contains an important variable,
   *                    or false otherwise.
   */
  setComponentImportant: function(name, isimportant)
  {
    if (this.components[name].important == isimportant)
      return;
    this.components[name].important = isimportant;

    if (this.importantComponents == 0 && isimportant)
      this.modlev.setImportImportant(this.number, true);
    else if (this.importantComponents == 1 && !isimportant)
      this.modlev.setImportImportant(this.number, false);

    if (isimportant)
      this.rcimportantt.AppendElement(this.components[name].resource);
    else
      this.rcimportantt.RemoveElement(this.components[name].resource, true);

    if (isimportant)
      this.importantComponents++;
    else
      this.importantComponents--;
  },

  stringifyImportantState: function()
  {
    var str = "";
    var k;
    for (k in this.components)
      if (this.components[k].important)
      {
        str += k + " " + this.components[k].stringifyImportantState() +
               ",";
      }
    return str;
  },

  /**
   * Destroys this ComponentLevel.
   */
  destroy: function()
  {
    var x;
    for (x in this.components)
    {
      this.components[x].destroy();
      this.components[x] = null;
    }
  }
};

/**
 * The prototype for the ModelLevel class.
 * ModelLevel classes represent the model level of the model=>import=>
 * component=>variable hierarachy, and maintain RDF to mirror this information.
 * @param model The underlying ModelInstance.
 * @param rdfserv The nsIRDFService.
 * @param rsallt The nsIRDFSource for the hierarchical view of all variables.
 * @param rsimportantt The nsIRDFSource for the hierarchical view of important variables.
 * @param rsallf The nsIRDFSource for the flat view of all variables.
 * @param rsimportantf The nsIRDFSource for the flat view of important variables.
 * @param rdfcu An nsIRDFContainerUtilities instance.
 * Side-effect This code will automatically add itself into the appropriate
 *             RDF containers.
 * Note This differs from ModelInstance because there is only one ModelInstance
 *      for a given set of initial conditions, but there is one ModelLevel per
 *      window that is viewing the variable hierarchy.
 */
function ModelLevel(model, rdfserv, rsallt, rsimportantt, rsallf, rsimportantf, rdfcu)
{
  this.model = model;
  this.rdfserv = rdfserv;
  this.rsallt = rsallt;
  this.rsimportantt = rsimportantt;
  this.rsallf = rsallf;
  this.rsimportantf = rsimportantf;
  this.rdfcu = rdfcu;
  this.imports = [];

  this.resource = window.rdfserv.GetResource
                  ("http://www.physiome.org/rdf/mozCellML#all-variables");
  this.rcallt = this.rdfcu.MakeSeq(this.rsallt, this.resource);
  this.rcimportantt = this.rdfcu.MakeSeq(this.rsimportantt, this.resource);
  this.rcallf = this.rdfcu.MakeSeq(this.rsallf, this.resource);
  this.rcimportantf = this.rdfcu.MakeSeq(this.rsimportantf, this.resource);
}

/**
 * The prototype for the ModelLevel class.
 */
ModelLevel.prototype =
{
  /**
   * An array of all imports.
   * Readonly outside of the ModelLevel class.
   */
  imports: [],

  /**
   * Find an ImportLevel, and create/add it if not found.
   * @param number The import number to add or find.
   * @return The existing or new ImportLevel.
   */
  addOrFindImport: function(number)
  {
    if (this.imports[number] == null)
    {
      this.imports[number] = new ImportLevel(number, this);
      this.rcallt.AppendElement(this.imports[number].resource);
    }
    return this.imports[number];
  },

  /**
   * Sets an import as containing(or not containing) an important variable.
   * @param number The number of the import to modify.
   * @param important True if the import contains one or more important
   *                  variables, or false otherwise.
   * Note You should call VariableLevel::setImportant(important) instead of
   *      calling this directly.
   */
  setImportImportant: function(number, important)
  {
    if (this.imports[number].isimportant == important)
      return;

    this.imports[number].isimportant = important;
    
    if (important)
      this.rcimportantt.AppendElement(this.imports[number].resource);
    else
      this.rcimportantt.RemoveElement(this.imports[number].resource, true);
  },

  /**
   * Converts the list of "important" variables into a string.
   */
  stringifyImportantState: function()
  {
    var str = "";
    var i;
    for (i = 0; i < this.imports.length; i++)
    {
      if (this.imports[i].isimportant)
        str += i + "," + this.imports[i].stringifyImportantState() + ";";
    }
    return str;
  },

  /**
   * Restores the list of "important" variables from a string from stringifyImportantState
   */
  restoreImportantState: function(impstate)
  {
    var spl1 = impstate.split(";");
    var i, j, k;
    for (i = 0; i < spl1.length; i++)
    {
      var spl2 = spl1[i].split(",");
      if (spl2.length == 0)
        continue;
      var impno = spl2[0];
      for (j = 1; j < spl2.length; j++)
      {
        var spl3 = spl2[j].split(" ");
        if (spl3.length == 0)
          continue;
        var compname = spl3[0];
        for (k = 1; k < spl3.length; k++)
        {
          if (spl3.length == 0)
            continue;
          var varname = spl3[k];
          if (varname == "")
            continue;
          this.imports[impno].components[compname].variables[varname].
               setImportant(true);
        }
      }
    }
  },
  
  /**
   * Destroys this ModelLevel.
   */
  destroy: function()
  {
    var i;
    for (i = 0; i < this.imports.length; i++)
    {
      this.imports[i].destroy();
      this.imports[i] = null;
    }
  }
};

function Graph(name, id)
{
  this.name = name;
  this.id = id;
  this.last_trace_id = 0;
  this.traces = [];
}

const COMPATIBILITY_NOAXIS=0;
const COMPATIBILITY_XAXIS=1;
const COMPATIBILITY_YAXIS=2;
const COMPATIBILITY_BOTHAXES=3;

Graph.prototype =
{
  /**
   * Adds a trace to the graph.
   * @param xvarlev The variable level to draw on the X axis.
   * @param yvarlev The variable level to draw on the Y axis.
   */
  addTrace: function(xvarlev, yvarlev)
  {
    traces.append(new GraphTrace(xvarlev, yvarlev));
  },

  /**
   * Gets the axis compatibility of a variable level. A variable is compatible
   * for an axis if the variable on the same axis of the existing traces have
   * the same units as the variable.
   * @param varlev The variable leve to test for compatibility.
   * @param cutoff The cut-off for how many traces are allowed before
   *               constraints take effect.
   * @return An integer bitfield of COMPATIBILITY_XAXIS and COMPATIBILITY_YAXIS.
   */
  getAxisCompatibility: function(varlev, cutoff)
  {
    if (varlev.cellmlvar.type == Components.interfaces.mozICellMLVariable.CONSTANT)
      return COMPATIBILITY_NOAXIS;
    if (this.traces.length <= cutoff)
    {
      // If there are no traces yet, there are no constraints... Also, if there
      // is only one trace, we need to let the user change it.
      return COMPATIBILITY_BOTHAXES;
    }
    var compat = 0;
    var cxunits = this.traces[0].xvarlev.cellmlvar.units;
    var cyunits = this.traces[0].yvarlev.cellmlvar.units;
    var myunits = varlev.cellmlvar.units;
    if (myunits == cxunits)
      compat = compat | COMPATIBILITY_XAXIS;
    if (myunits == cyunits)
      compat = compat | COMPATIBILITY_YAXIS;
    return compat;
  },

  /**
   * Updates compatibility for an entire axis ModelLevel.
   * @param modlev The model level to update.
   * @param mask The axis mask applicable to this model level.
   * @param cutoff The cut-off for how many traces are allowed before
   *               constraints take effect.
   * Note: All bits in mask must be set on the axis compatibility in order for
   *       a variable to be flagged as important.
   */
  updateAxisCompatibility: function(modlev, mask, cutoff)
  {
    var imp, comp, varn;
    for (imp in modlev.imports)
      for (comp in modlev.imports[imp].components)
        for (varn in modlev.imports[imp].components[comp].variables)
        {
          var v = modlev.imports[imp].components[comp].variables[varn];
          v.setImportant((this.getAxisCompatibility(v, cutoff) & mask) == mask);
        }
  }
};

/**
 * The constructor for a graph trace.
 * @param xvarlev The VariableLevel for the variable to draw on the X axis.
 * @param yvarlev The VariableLevel for the variable to draw on the Y axis.
 */
function GraphTrace(xvarlev, yvarlev)
{
  this.xvarlev = xvarlev;
  this.yvarlev = yvarlev;
  this.colour = parseInt(Math.random()*(1<<24));
}

GraphTrace.prototype =
{
  assignToColourPicker: function(cp)
  {
    var h = "0123456789abcdef";
    var cl = "#" +
      h[((this.colour & 0xF00000) >> 20)] +
      h[((this.colour & 0xF0000) >> 16)] +
      h[((this.colour & 0xF000) >> 12)] +
      h[((this.colour & 0xF00) >> 8)] +
      h[((this.colour & 0xF0) >> 4)] +
      h[(this.colour & 0xF)];
    cp.color = cl;
    var cpb = cp.ownerDocument.getAnonymousElementByAttribute(cp, "anonid", "colorbox");
    if (cpb != null)
      cpb.style.backgroundColor = cl;
  },

  parseColour: function(col)
  {
    var uh = {"0":0,"1": 1,"2": 2,"3": 3,"4": 4,"5": 5,"6": 6,"7":7,"8":8,
              "9":9,"A":10,"B":11,"C":12,"D":13,"E":14,"F":15};
    this.colour = (uh[col[1]] << 20) + (uh[col[2]] << 16) + (uh[col[3]] << 12) +
      (uh[col[4]] << 8) + (uh[col[5]] << 4) + uh[col[6]];
  }
};
