/*	-*- C++ -*-
 * Copyright (C) 2001,  Christof Meerwald <cmeerw@web.de>
 * http://xpcom2corba.sunsite.dk/
 */

/*
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License
 * at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and
 * limitations under the License.
 *
 * The Original Code is the XPCOM-CORBA Bridge package.
 *
 * The Initial Developer of the Original Code is Christof Meerwald.
 * Portions created by Christof Meerwald are Copyright (C) 2001
 * Christof Meerwald.  All Rights Reserved.
 *
 * Contributor(s): Christof Meerwald <cmeerw@web.de> (original author)
 */

#include "corba_poa.h"
#include "corba_proxy.h"
#include "corba_servant.h"

#include <cstring>

// TODO: debugging
#include <iostream>


CORBA_Proxy_Factory::Factory_Map *CORBA_Proxy_Factory::factories_ = 0;


CORBA_Proxy_Factory::CORBA_Proxy_Factory(const char *repo_id,
					 const CORBA_Proxy_Factory *factory)
{
  // Hmm, static initialization doesn't work, therefore we have to use
  // a singleton here
  if (factories_ == 0)
    factories_ = new Factory_Map();

  factory_iter_ =
    factories_->insert(Factory_Map::value_type(repo_id, factory)).first;
}

CORBA_Proxy_Factory::~CORBA_Proxy_Factory()
{
  factories_->erase(factory_iter_);

  // the last one also deletes the map
  if (factories_->size() == 0)
  {
    delete factories_;
    factories_ = 0;
  }
}


CORBA_Proxy_Servant *
CORBA_Proxy_Factory::GetProxy(const char *repo_id, nsISupports *obj)
{
  // Hmm, static initialization doesn't work, therefore we have to use
  // a singleton here
  if (factories_ == 0)
    factories_ = new Factory_Map();

  Factory_Map::const_iterator iter = factories_->find(repo_id);
  if (iter != factories_->end())
  {
    return iter->second->GetProxy(obj);
  }

  return 0;
}


CORBA_Proxy_Object::CORBA_Proxy_Object(nsISupports *xpcom_obj)
  : xpcom_obj_(xpcom_obj), ref_count_(1)
{
}


CORBA_Proxy_Object::~CORBA_Proxy_Object()
{
}


CORBA::Boolean CORBA_Proxy_Object::_is_a(const char *type)
{
  RepoId_Map::const_iterator repo_iter = repo_ids_.find(type);

  if (repo_iter == repo_ids_.end())
  {
    return add_repoid(type);
  }
  else
  {
    return true;
  }

  return false;
}


CORBA::Boolean CORBA_Proxy_Object::_non_existent()
{
  return false;
}


void CORBA_Proxy_Object::add_ref()
{
  ++ref_count_;
}


void CORBA_Proxy_Object::release_ref()
{
  if (--ref_count_ <= 0)
  {
    // unregister proxy object
    the_corba_poa->unregister_object(oid_);
    oid_ = NULL;

    // TODO: delete proxy object
    // assume for now that no other operation is in progress
    delete this;
  }
}


bool CORBA_Proxy_Object::add_repoid(const char *type)
{
  CORBA_Proxy_Servant *proxy_servant =
    CORBA_Proxy_Factory::GetProxy(type, xpcom_obj_);

  if (proxy_servant)
  {
    repo_ids_[proxy_servant->_repository_id()] = proxy_servant;
    add_operations(proxy_servant->_operation_names(), proxy_servant);

    add_operations_recursive(proxy_servant->_base_repository_ids());

    return true;
  }

  return false;
}


void CORBA_Proxy_Object::oid(const char *oid_str)
{
  oid_ = oid_str;
}


void CORBA_Proxy_Object::add_operations(const char **operation_iter,
					CORBA_Proxy_Servant *proxy_servant)
{
  while (*operation_iter)
  {
    bool replaced =
      operations_.insert(Operation_Map::value_type(*operation_iter,
						   proxy_servant)).second;
    if (replaced)
    {
      // TODO: oh dear
    }

    operation_iter++;
  }
}


void CORBA_Proxy_Object::add_operations_recursive(const char **repo_iter)
{
  // check base interfaces
  while (*repo_iter)
  {
    CORBA_Proxy_Servant *&proxy_servant = repo_ids_[*repo_iter];
    if (!proxy_servant)
    {
      proxy_servant = CORBA_Proxy_Factory::GetProxy(*repo_iter, xpcom_obj_);
      if (proxy_servant)
      {
	add_operations(proxy_servant->_operation_names(), proxy_servant);

	add_operations_recursive(proxy_servant->_base_repository_ids());
      }
      else
      {
	// TODO: hmm...
      }
    }

    repo_iter++;
  }
}


PortableServer::Servant
CORBA_Proxy_Object::servant_for_operation(const char *operation)
{
  const Operation_Map::const_iterator operation_iter =
    operations_.find(operation);

  if (operation_iter != operations_.end())
  {
    return operation_iter->second;
  }


  // hmm, operation not found... :-)
  std::cerr << "CORBA_Proxy_Object::servant_for_operation not found"
	    << std::endl;

  // print out some debugging information

  for (Operation_Map::const_iterator oiter = operations_.begin();
       oiter != operations_.end();
       ++oiter)
  {
    std::cerr << oiter->first << std::endl;
  }

  std::cerr << "registered repository ids" << std::endl;
  for (RepoId_Map::const_iterator riter = repo_ids_.begin();
       riter != repo_ids_.end();
       ++riter)
  {
    std::cerr << riter->first << std::endl;
  }


  return NULL;
}
